﻿using RootPro.RootProCAD;
using RootPro.RootProCAD.Command;
using RootPro.RootProCAD.Geometry;
using RootPro.RootProCAD.UI;
using System;
using System.IO;
using System.Reflection;
using System.Threading;
using System.Windows.Forms;
using System.Xml;
using System.Xml.Serialization;

namespace RCAddInMeasureTransformedCoordsCS
{
    [System.AddIn.AddIn("AppAddIn", Version = "1.0", Publisher = "", Description = "")]
    public partial class AppAddIn
    {
        // アドイン設定
        AddInSettings settings = new AddInSettings();

        // コマンド名
        string MeasurTrsfmCoordsCommandName = "RootPro.AddIn.MeasureTrsfmCoordsCommand";

        private void AppAddIn_Startup(object sender, EventArgs e)
        {
            // 多言語対応リソースのDLLファイルをロード
            AppDomain.CurrentDomain.AssemblyResolve += new ResolveEventHandler(ResolveAssembly);

            // マクロコマンド追加
            CommandManager.AddMacroCommand("RCAddInMeasureTransformedCoordsCS", Properties.Resources.CommandName, MacroCommand);

            // 変換座標測定コマンドを追加する
            Command command = CommandManager.AddCommand(MeasurTrsfmCoordsCommandName, Properties.Resources.CommandName);

            // 基準点 1 パラメータ追加
            command.ParameterGroups.Add(ParameterItemType.Point, 0, Properties.Resources.BasedPoint + "1");
            PointParameterItem pointParameterItem = (PointParameterItem)command.GetParameterItemByID(0);
            pointParameterItem.Description = Properties.Resources.BasedPointParameterDescription;

            // 基準点 2 パラメータ追加
            command.ParameterGroups.Add(ParameterItemType.Point, 1, Properties.Resources.BasedPoint + "2");
            pointParameterItem = (PointParameterItem)command.GetParameterItemByID(1);
            pointParameterItem.Description = Properties.Resources.BasedPointParameterDescription;

            // 変換点 1 パラメータ追加
            command.ParameterGroups.Add(ParameterItemType.Point, 2, Properties.Resources.ConvertedPoint + "1");
            pointParameterItem = (PointParameterItem)command.GetParameterItemByID(2);
            pointParameterItem.Description = Properties.Resources.ConvertedPointParameterDescription;

            // 変換点 2 パラメータ追加
            command.ParameterGroups.Add(ParameterItemType.Point, 3, Properties.Resources.ConvertedPoint + "2");
            pointParameterItem = (PointParameterItem)command.GetParameterItemByID(3);
            pointParameterItem.Description = Properties.Resources.ConvertedPointParameterDescription;

            // 測定点 パラメータ追加
            command.ParameterGroups.Add(ParameterItemType.Point, 4, Properties.Resources.MeasuringPoint);
            pointParameterItem = (PointParameterItem)command.GetParameterItemByID(4);
            pointParameterItem.Description = Properties.Resources.MeasuringPointParameterDescription;

            //開始イベントと終了イベントの追加
            command.Begin += new CommandBeginEventHandler(command_Begin);
            command.End += new EventHandler(command_End);
        }

        private void AppAddIn_Shutdown(object sender, EventArgs e)
        {
            CommandManager.RemoveCommand(MeasurTrsfmCoordsCommandName);

            // マクロコマンドの削除
            CommandManager.RemoveMacroCommand(MacroCommand);
        }

        #region VSTA generated code
        private void InternalStartup()
        {
            this.Startup += new System.EventHandler(AppAddIn_Startup);
            this.Shutdown += new System.EventHandler(AppAddIn_Shutdown);
        }
        #endregion

        private Assembly ResolveAssembly(object sender, ResolveEventArgs args)
        {
            AssemblyName assemblyName = new AssemblyName(args.Name);

            foreach(Assembly assembly in AppDomain.CurrentDomain.GetAssemblies())
            {
                if(AssemblyName.ReferenceMatchesDefinition(assemblyName, assembly.GetName()))
                    return assembly;
            }

            if(assemblyName.Name.EndsWith(".resources"))
            {
                if(assemblyName.CultureName.Equals(Thread.CurrentThread.CurrentUICulture.Name, StringComparison.OrdinalIgnoreCase))
                {
                    string dllPath = Path.GetDirectoryName(this.AddInFullPath);
                    dllPath = Path.Combine(dllPath, assemblyName.CultureName);
                    dllPath = Path.Combine(dllPath, assemblyName.Name + ".dll");

                    if(File.Exists(dllPath))
                    {
                        try
                        {
                            return Assembly.LoadFrom(dllPath);
                        }
                        catch
                        {
                        }
                    }
                }
            }

            return null;
        }

        private void MacroCommand()
        {
            CommandManager.ExecuteCommand(MeasurTrsfmCoordsCommandName);
        }

        /// <summary>
        /// コマンドの開始時に呼ばれる
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void command_Begin(Object sender, CommandBeginEventArgs e)
        {
            Command command = (Command)sender;

            // 前回のパラメータを読み込み
            LoadSettings();

            // コマンドパラメータを更新
            UpdateParameters();

            // パラメータ変更イベントの追加
            command.ParameterChanged += new CommandParameterChangedEventHandler(command_ParameterChanged);
        }

        /// <summary>
        /// コマンドの終了時に呼ばれる
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void command_End(Object sender, EventArgs e)
        {
            Command command = (Command)sender;

            // パラメータ変更イベントの削除
            command.ParameterChanged -= new CommandParameterChangedEventHandler(command_ParameterChanged);

            // パラメータを保存
            SaveSettings();
        }

        /// <summary>
        /// パラメータ変更イベント
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void command_ParameterChanged(Object sender, CommandParameterChangedEventArgs e)
        {
            Command command = (Command)sender;

            bool bReturn = false;

            // 基準点
            Point2d basedPoint1 = new Point2d(0.0, 0.0);
            Point2d basedPoint2 = new Point2d(0.0, 0.0);
            // 変換点
            Point2d convertedPoint1 = new Point2d(0.0, 0.0);
            Point2d convertedPoint2 = new Point2d(0.0, 0.0);

            // 基準点 1 の取得
            PointParameterItem pointParameterItem = (PointParameterItem)command.GetParameterItemByID(0);
            if(!pointParameterItem.IsEmpty)
            {
                settings.basedPoint1X = basedPoint1.X = pointParameterItem.Point.X;
                settings.basedPoint1Y = basedPoint1.Y = pointParameterItem.Point.Y;
            }
            else
            {
                bReturn = true;
                // 空にした場合は設定を初期値に戻す
                if(e.ParameterID == 0)
                {
                    settings.basedPoint1X = double.MaxValue;
                    settings.basedPoint1Y = double.MaxValue;
                }
            }

            // 基準点 2 の取得
            pointParameterItem = (PointParameterItem)command.GetParameterItemByID(1);
            if(!pointParameterItem.IsEmpty)
            {
                settings.basedPoint2X = basedPoint2.X = pointParameterItem.Point.X;
                settings.basedPoint2Y = basedPoint2.Y = pointParameterItem.Point.Y;
            }
            else
            {
                bReturn = true;
                // 空にした場合は設定を初期値に戻す
                if(e.ParameterID == 1)
                {
                    settings.basedPoint2X = double.MaxValue;
                    settings.basedPoint2Y = double.MaxValue;
                }
            }

            // 変換点 1 の取得
            pointParameterItem = (PointParameterItem)command.GetParameterItemByID(2);
            if(!pointParameterItem.IsEmpty)
            {
                settings.convertedPoint1X = convertedPoint1.X = pointParameterItem.Point.X;
                settings.convertedPoint1Y = convertedPoint1.Y = pointParameterItem.Point.Y;
            }
            else
            {
                bReturn = true;
                // 空にした場合は設定を初期値に戻す
                if(e.ParameterID == 2)
                {
                    settings.convertedPoint1X = double.MaxValue;
                    settings.convertedPoint1Y = double.MaxValue;
                }
            }
            // 変換点 2 の取得
            pointParameterItem = (PointParameterItem)command.GetParameterItemByID(3);
            if(!pointParameterItem.IsEmpty)
            {
                settings.convertedPoint2X = convertedPoint2.X = pointParameterItem.Point.X;
                settings.convertedPoint2Y = convertedPoint2.Y = pointParameterItem.Point.Y;
            }
            else
            {
                bReturn = true;
                // 空にした場合は設定を初期値に戻す
                if(e.ParameterID == 3)
                {
                    settings.convertedPoint2X = double.MaxValue;
                    settings.convertedPoint2Y = double.MaxValue;
                }
            }

            // 測定点 の取得
            pointParameterItem = (PointParameterItem)command.GetParameterItemByID(4);
            if(pointParameterItem.IsEmpty)
                return;

            // 基準点 1,2、変換点 1,2 が未入力の場合
            if(bReturn)
            {
                // 先に測定点を入力することは不可とする
                if(e.ParameterID == 4)
                {
                    command.MoveToNextParameter();
                    e.CommandState = CommandState.ParameterError;
                }
                return;
            }

            // 測定点
            Point2d measuringPoint = new Point2d(pointParameterItem.Point.X, pointParameterItem.Point.Y);

            // 基準点間の距離
            double basedPointDistance = Geometry.GetDistance(basedPoint1, basedPoint2);

            // 変換点間の距離
            double convertedPointDistance = Geometry.GetDistance(convertedPoint1, convertedPoint2);

            // 基準点間の角度
            double basedPointAngle = Geometry.GetAngle(basedPoint1, basedPoint2);

            // 変換点間の角度
            double convertedPointAngle = Geometry.GetAngle(convertedPoint1, convertedPoint2);

            // A -> B を A'-> B' の座標値としたとき、AB 座標系の C をA'B'座標系に変換した場合の C'を求める
            var pointMarker = ShapeFactory.CreatePointMarker(measuringPoint);
            // Transform(A, A'B'長/AB長, A'B'長/AB長, A'B'角-AB角, A')
            pointMarker.Transform(basedPoint1, 
                convertedPointDistance / basedPointDistance, 
                convertedPointDistance / basedPointDistance,
                convertedPointAngle - basedPointAngle,
                convertedPoint1);

            // 変換した点の座標値をクリップボードにコピーする
            string convertedPtStr = NumericTextConverter.PointToText(pointMarker.Point, LengthUnitFormatType.Decimal, ActiveDocument.LengthPrecision, true);
            Clipboard.SetText(convertedPtStr);

            string statusBarMsg = Properties.Resources.ClipboardMessage + " (" + convertedPtStr + ")";
            StatusBar.SetInformation(statusBarMsg, 5000);

            // コマンドを継続するためにパラメータエラーをセットする
            e.CommandState = CommandState.ParameterError;

            return;
        }

        /// <summary>
        /// アドインの設定値クラス
        /// </summary>
        public class AddInSettings
        {
            // 基準点 1 X
            public double basedPoint1X { get; set; }

            // 基準点 1 Y
            public double basedPoint1Y { get; set; }

            // 基準点 2 X
            public double basedPoint2X { get; set; }

            // 基準点 2 Y
            public double basedPoint2Y { get; set; }

            // 変換点 1 X
            public double convertedPoint1X { get; set; }

            // 変換点 1 Y
            public double convertedPoint1Y { get; set; }

            // 変換点 2 X
            public double convertedPoint2X { get; set; }

            // 変換点 2 Y
            public double convertedPoint2Y { get; set; }

            /// <summary>
            /// コンストラクタ―
            /// </summary>
            public AddInSettings()
            {
                Clear();
            }

            /// <summary>
            /// 初期化
            /// </summary>
            public void Clear()
            {
                basedPoint1X = 0.0;
                basedPoint1Y = double.MaxValue;
                basedPoint2X = double.MaxValue;
                basedPoint2Y = double.MaxValue;
                convertedPoint1X = double.MaxValue;
                convertedPoint1Y = double.MaxValue;
                convertedPoint2X = double.MaxValue;
                convertedPoint2Y = double.MaxValue;
            }
        }

        /// <summary>
        /// 設定ファイルパスを取得
        /// </summary>
        public string GetSettinsFileName()
        {
            // アドインフォルダ
            return Path.GetDirectoryName(AddInFullPath) + "\\settings.xml";
        }


        /// <summary>
        /// 設定の読み込み
        /// </summary>
        public void LoadSettings()
        {
            string settingsFileName = GetSettinsFileName();
            if(!File.Exists(settingsFileName))
                return;

            try
            {
                // XMLファイルから設定を読み込む
                using(var reader = XmlReader.Create(settingsFileName))
                {
                    var serializer = new XmlSerializer(typeof(AddInSettings));
                    settings = (AddInSettings)serializer.Deserialize(reader);
                }
            }
            catch(Exception)
            {
                return;
            }
        }

        /// <summary>
        /// 設定の保存
        /// </summary>
        public void SaveSettings()
        {
            // 設定をXMLファイルに保存
            var serializer = new XmlSerializer(typeof(AddInSettings));
            using(var writer = new StreamWriter(GetSettinsFileName()))
                serializer.Serialize(writer, settings);
        }

        /// <summary>
        /// コマンドパラメータを更新する
        /// </summary>
        public void UpdateParameters()
        {
            PointParameterItem pointParameterItem = null;
            bool mooveNext = false;

            // アドインコマンド取得
            Command command = CommandManager.Commands[MeasurTrsfmCoordsCommandName];

            // 基準点 1
            if(settings.basedPoint1X != double.MaxValue && settings.basedPoint1Y != double.MaxValue)
            {
                pointParameterItem = (PointParameterItem)command.GetParameterItemByID(0);
                pointParameterItem.Point = new Point2d(settings.basedPoint1X, settings.basedPoint1Y);
                mooveNext = true;
            }
            // 基準点 2
            if(settings.basedPoint2X != double.MaxValue && settings.basedPoint2Y != double.MaxValue)
            {
                pointParameterItem = (PointParameterItem)command.GetParameterItemByID(1);
                pointParameterItem.Point = new Point2d(settings.basedPoint2X, settings.basedPoint2Y);
            }
            // 変換点 1
            if(settings.convertedPoint1X != double.MaxValue && settings.convertedPoint1Y != double.MaxValue)
            {
                pointParameterItem = (PointParameterItem)command.GetParameterItemByID(2);
                pointParameterItem.Point = new Point2d(settings.convertedPoint1X, settings.convertedPoint1Y);
            }
            // 変換点 2
            if(settings.convertedPoint2X != double.MaxValue && settings.convertedPoint2Y != double.MaxValue)
            {
                pointParameterItem = (PointParameterItem)command.GetParameterItemByID(3);
                pointParameterItem.Point = new Point2d(settings.convertedPoint2X, settings.convertedPoint2Y);
            }

            if(mooveNext)
                command.MoveToNextParameter();
        }
    }
}
